import fsspec
import numpy as np
import pandas as pd
import pytest
import xarray as xr
import fsspec.implementations.reference as reffs
import ujson

import kerchunk.combine
import kerchunk.zarr
import kerchunk.utils


@pytest.fixture(scope="module")
def ds():
    ds = xr.Dataset(
        {
            "x": xr.DataArray(np.linspace(-np.pi, np.pi, 10), dims=["x"]),
            "y": xr.DataArray(np.linspace(-np.pi / 2, np.pi / 2, 10), dims=["y"]),
            "time": xr.DataArray(pd.date_range("2020", "2021"), dims=["time"]),
        },
    )
    ds["temp"] = (
        np.cos(ds.x)
        * np.sin(ds.y)
        * xr.ones_like(ds.time).astype("float")
        * np.random.random(ds.time.shape)
    )
    return ds


@pytest.fixture
def zarr_in_zip(tmpdir, ds):
    def _zip(file):
        import os
        import zipfile

        filename = file + os.path.extsep + "zip"
        with zipfile.ZipFile(
            filename, "w", compression=zipfile.ZIP_STORED, allowZip64=True
        ) as fh:
            for root, _, filenames in os.walk(file):
                for each_filename in filenames:
                    each_filename = os.path.join(root, each_filename)
                    fh.write(each_filename, os.path.relpath(each_filename, file))
        return filename

    fn = f"{tmpdir}/test.zarr"
    ds.to_zarr(fn, mode="w", zarr_format=2)
    return _zip(fn)


def test_zarr_in_zip(zarr_in_zip, ds):
    out = kerchunk.zarr.ZarrToZarr(
        url="zip://", storage_options={"fo": zarr_in_zip}
    ).translate()

    ds2 = xr.open_dataset(
        out,
        engine="kerchunk",
        backend_kwargs={
            "storage_options": {
                "remote_protocol": "zip",
                "remote_options": {"fo": zarr_in_zip},
            }
        },
    )
    assert ds.equals(ds2)

    # tests inlining of metadata
    fs = fsspec.filesystem(
        "reference", fo=out, remote_protocol="zip", remote_options={"fo": zarr_in_zip}
    )
    assert isinstance(fs.references["temp/.zarray"], (str, bytes))


def test_zarr_combine(tmpdir, ds):
    fn1 = f"{tmpdir}/test1.zarr"
    ds.to_zarr(fn1, zarr_format=2)

    one = kerchunk.zarr.ZarrToZarr(fn1, inline_threshold=0).translate()
    fn = f"{tmpdir}/out.parq"
    out = reffs.LazyReferenceMapper.create(fn)
    mzz = kerchunk.combine.MultiZarrToZarr([one], concat_dims=["time"], out=out)
    mzz.translate()

    ds2 = xr.open_dataset(fn, engine="kerchunk")
    assert ds.equals(ds2)


def test_zarr_json_dump_succeeds(tmpdir, ds):
    fn1 = f"{tmpdir}/test1.zarr"
    ds.to_zarr(fn1, zarr_format=2)

    one = kerchunk.zarr.ZarrToZarr(
        fn1,
        inline_threshold=0,
    ).translate()
    ujson.dumps(one)



import asyncio
import fsspec
import json
import os.path as osp
import hdf5plugin
import zarr.core
import zarr.core.buffer
import zarr.core.group

import kerchunk.hdf
import numpy as np
import pytest
import xarray as xr
import zarr

from fsspec.implementations.asyn_wrapper import AsyncFileSystemWrapper
from kerchunk.hdf import SingleHdf5ToZarr, has_visititems_links
from kerchunk.combine import MultiZarrToZarr, drop
from kerchunk.utils import fs_as_store, refs_as_fs, refs_as_store

here = osp.dirname(__file__)


def test_single():
    """Test creating references for a single HDF file"""
    url = "s3://noaa-nwm-retro-v2.0-pds/full_physics/2017/201704010000.CHRTOUT_DOMAIN1.comp"
    so = dict(anon=True, default_fill_cache=False, default_cache_type="none")

    with fsspec.open(url, **so) as f:
        h5chunks = SingleHdf5ToZarr(f, url, storage_options=so, inline_threshold=1)
        test_dict = h5chunks.translate()

    with open("test_dict.json", "w") as f:
        json.dump(test_dict, f)

    store = refs_as_store(test_dict, remote_options=dict(asynchronous=True, anon=True))
    ds = xr.open_zarr(store, zarr_format=2, consolidated=False)

    with fsspec.open(url, **so) as f:
        expected = xr.open_dataset(f, engine="h5netcdf")
        xr.testing.assert_equal(ds.drop_vars("crs"), expected.drop_vars("crs"))


def test_single_direct_open():
    """Test creating references by passing the url directly to SingleHdf5ToZarr for a single HDF file"""
    url = "s3://noaa-nwm-retro-v2.0-pds/full_physics/2017/201704010000.CHRTOUT_DOMAIN1.comp"
    so = dict(anon=True, default_fill_cache=False, default_cache_type="first")

    test_dict = SingleHdf5ToZarr(
        h5f=url, inline_threshold=300, storage_options=so
    ).translate()

    store = refs_as_store(test_dict, remote_options=dict(asynchronous=True, anon=True))

    ds_direct = xr.open_dataset(
        store, engine="zarr", zarr_format=2, backend_kwargs=dict(consolidated=False)
    )

    with fsspec.open(url, **so) as f:
        h5chunks = SingleHdf5ToZarr(f, url, storage_options=so)
        test_dict = h5chunks.translate()

    store = refs_as_store(test_dict, remote_options=dict(asynchronous=True, anon=True))

    ds_from_file_opener = xr.open_dataset(
        store, engine="zarr", zarr_format=2, backend_kwargs=dict(consolidated=False)
    )

    xr.testing.assert_equal(
        ds_from_file_opener.drop_vars("crs"), ds_direct.drop_vars("crs")
    )


urls = [
    "s3://" + p
    for p in [
        "noaa-nwm-retro-v2.0-pds/full_physics/2017/201704010000.CHRTOUT_DOMAIN1.comp",
        "noaa-nwm-retro-v2.0-pds/full_physics/2017/201704010100.CHRTOUT_DOMAIN1.comp",
        "noaa-nwm-retro-v2.0-pds/full_physics/2017/201704010200.CHRTOUT_DOMAIN1.comp",
    ]
]
so = dict(anon=True, default_fill_cache=False, default_cache_type="first")


def test_multizarr(generate_mzz):
    """Test creating a combined reference file with MultiZarrToZarr"""
    mzz = generate_mzz
    test_dict = mzz.translate()

    store = refs_as_store(test_dict, remote_options=dict(asynchronous=True, anon=True))
    ds = xr.open_dataset(
        store, engine="zarr", zarr_format=2, backend_kwargs=dict(consolidated=False)
    )

    with fsspec.open_files(urls, **so) as fs:
        expts = [xr.open_dataset(f, engine="h5netcdf") for f in fs]
        expected = xr.concat(expts, dim="time")

        assert set(ds) == set(expected)
        for name in ds:
            exp = {
                k: (
                    (v.tolist() if v.size > 1 else v[0])
                    if isinstance(v, np.ndarray)
                    else v
                )
                for k, v in expected[name].attrs.items()
            }
            assert dict(ds[name].attrs) == dict(exp)
        for coo in ds.coords:
            assert (ds[coo].values == expected[coo].values).all()


@pytest.fixture(scope="module")
def generate_mzz():
    """This function generates a MultiZarrToZarr class for tests"""

    dict_list = []

    for u in urls:
        with fsspec.open(u, **so) as inf:
            h5chunks = SingleHdf5ToZarr(inf, u, inline_threshold=100)
            dict_list.append(h5chunks.translate())

    mzz = MultiZarrToZarr(
        dict_list,
        remote_protocol="s3",
        remote_options={"anon": True},
        concat_dims=["time"],
        preprocess=drop("reference_time"),
    )
    return mzz


@pytest.fixture()
def times_data(tmpdir):
    lat = xr.DataArray(np.linspace(-90, 90, 10), dims=["lat"], name="lat")
    lon = xr.DataArray(np.linspace(-90, 90, 10), dims=["lon"], name="lon")
    time_attrs = {"axis": "T", "long_name": "time", "standard_name": "time"}
    time1 = xr.DataArray(
        np.arange(-631108800000000000, -630158390000000000, 86400000000000).view(
            "datetime64[ns]"
        ),
        dims=["time"],
        name="time",
        attrs=time_attrs,
    )

    x1 = xr.DataArray(
        np.zeros((12, 10, 10)),
        dims=["time", "lat", "lon"],
        coords={"time": time1, "lat": lat, "lon": lon},
        name="prcp",
    )
    url = str(tmpdir.join("x1.nc"))
    x1.to_netcdf(url, engine="h5netcdf")
    return x1, url


def test_times(times_data):
    x1, url = times_data
    # Test taken from https://github.com/fsspec/kerchunk/issues/115#issue-1091163872
    with fsspec.open(url) as f:
        h5chunks = SingleHdf5ToZarr(f, url)
        test_dict = h5chunks.translate()

    localfs = AsyncFileSystemWrapper(fsspec.filesystem("file"))
    store = refs_as_store(test_dict, fs=localfs)
    result = xr.open_dataset(
        store, engine="zarr", zarr_format=2, backend_kwargs=dict(consolidated=False)
    )
    expected = x1.to_dataset()
    xr.testing.assert_equal(result, expected)


def test_times_str(times_data):
    # as test above, but using str input for SingleHdf5ToZarr file
    x1, url = times_data
    # Test taken from https://github.com/fsspec/kerchunk/issues/115#issue-1091163872
    h5chunks = SingleHdf5ToZarr(url)
    test_dict = h5chunks.translate()

    localfs = AsyncFileSystemWrapper(fsspec.filesystem("file"))
    store = refs_as_store(test_dict, fs=localfs)
    result = xr.open_dataset(
        store, engine="zarr", zarr_format=2, backend_kwargs=dict(consolidated=False)
    )
    expected = x1.to_dataset()
    xr.testing.assert_equal(result, expected)


# https://stackoverflow.com/a/43935389/3821154
txt = "the change of water into water vapour"


def test_string_embed():
    fn = osp.join(here, "vlen.h5")
    h = kerchunk.hdf.SingleHdf5ToZarr(fn, fn, vlen_encode="embed")
    out = h.translate()

    localfs = AsyncFileSystemWrapper(fsspec.filesystem("file"))
    fs = refs_as_fs(out, fs=localfs)
    # assert txt in fs.references["vlen_str/0"]
    store = fs_as_store(fs)
    z = zarr.open(store, zarr_format=2)
    assert "String" in str(z["vlen_str"])
    assert z["vlen_str"][0] == txt
    assert (z["vlen_str"][1:] == "").all()


def test_string_pathlib():
    # essentially copied test above
    import pathlib

    fn = osp.join(here, "vlen.h5")
    h = kerchunk.hdf.SingleHdf5ToZarr(pathlib.Path(fn), vlen_encode="embed")
    out = h.translate()
    fs = fsspec.filesystem("reference", fo=out)
    assert txt in fs.references["vlen_str/0"]
    z = zarr.open(fs_as_store(fs))
    assert "String" in str(z["vlen_str"])
    assert z["vlen_str"][0] == txt
    assert (z["vlen_str"][1:] == "").all()


def test_string_null():
    fn = osp.join(here, "vlen.h5")
    h = kerchunk.hdf.SingleHdf5ToZarr(fn, fn, vlen_encode="null", inline_threshold=0)
    out = h.translate()
    localfs = AsyncFileSystemWrapper(fsspec.filesystem("file"))
    store = refs_as_store(out, fs=localfs)
    z = zarr.open(store, zarr_format=2)
    assert z["vlen_str"].dtype == "O"
    assert (z["vlen_str"][:] == None).all()


def test_string_leave():
    fn = osp.join(here, "vlen.h5")
    with open(fn, "rb") as f:
        h = kerchunk.hdf.SingleHdf5ToZarr(
            f, fn, vlen_encode="leave", inline_threshold=0
        )
        out = h.translate()

    localfs = AsyncFileSystemWrapper(fsspec.filesystem("file"))
    store = refs_as_store(out, fs=localfs)
    z = zarr.open(store, zarr_format=2)
    assert z["vlen_str"].dtype == "S16"
    assert z["vlen_str"][0]  # some obscured ID
    assert (z["vlen_str"][1:] == b"").all()


def test_string_decode():
    fn = osp.join(here, "vlen.h5")
    with open(fn, "rb") as f:
        h = kerchunk.hdf.SingleHdf5ToZarr(
            f, fn, vlen_encode="encode", inline_threshold=0
        )
        out = h.translate()
    localfs = AsyncFileSystemWrapper(fsspec.filesystem("file"))
    fs = refs_as_fs(out, fs=localfs)
    assert txt in fs.cat("vlen_str/.zarray").decode()  # stored in filter def
    store = fs_as_store(fs)
    z = zarr.open(store, zarr_format=2)
    assert z["vlen_str"][0] == txt
    assert (z["vlen_str"][1:] == "").all()


def test_compound_string_null():
    fn = osp.join(here, "vlen2.h5")
    with open(fn, "rb") as f:
        h = kerchunk.hdf.SingleHdf5ToZarr(f, fn, vlen_encode="null", inline_threshold=0)
        out = h.translate()
    localfs = AsyncFileSystemWrapper(fsspec.filesystem("file"))
    store = refs_as_store(out, fs=localfs)
    z = zarr.open(store, zarr_format=2)
    assert z["vlen_str"][0].tolist() == (10, None)
    assert (z["vlen_str"][1:]["ints"] == 0).all()
    assert (z["vlen_str"][1:]["strs"] == None).all()


def test_compound_string_leave():
    fn = osp.join(here, "vlen2.h5")
    with open(fn, "rb") as f:
        h = kerchunk.hdf.SingleHdf5ToZarr(
            f, fn, vlen_encode="leave", inline_threshold=0
        )
        out = h.translate()
    localfs = AsyncFileSystemWrapper(fsspec.filesystem("file"))
    store = refs_as_store(out, fs=localfs)
    z = zarr.open(store, zarr_format=2)
    assert z["vlen_str"][0]["ints"] == 10
    assert z["vlen_str"][0]["strs"]  # random ID
    assert (z["vlen_str"][1:]["ints"] == 0).all()
    assert (z["vlen_str"][1:]["strs"] == b"").all()


def test_compound_string_encode():
    fn = osp.join(here, "vlen2.h5")
    with open(fn, "rb") as f:
        h = kerchunk.hdf.SingleHdf5ToZarr(
            f, fn, vlen_encode="encode", inline_threshold=0, error="raise"
        )
        out = h.translate()
        print(json.dumps(out, indent=4))
    localfs = AsyncFileSystemWrapper(fsspec.filesystem("file"))
    store = refs_as_store(out, fs=localfs)
    z = zarr.open(store, zarr_format=2)
    assert z["vlen_str"][0]["ints"] == 10
    assert z["vlen_str"][0]["strs"] == "water"
    assert (z["vlen_str"][1:]["ints"] == 0).all()
    assert (z["vlen_str"][1:]["strs"] == "").all()


# def test_compact():
#     pytest.importorskip("ipfsspec")
#     h = kerchunk.hdf.SingleHdf5ToZarr(
#         "ipfs://QmVZc4TzRP7zydgKzDX7CH2JpYw2LJKkWBm6jhCfigeon6"
#     )
#     out = h.translate()
#
#     m = fsspec.get_mapper("reference://", fo=out)
#     g = zarr.open(m)
#     assert np.allclose(g.ancillary_data.atlas_sdp_gps_epoch[:], 1.19880002e09)
#
#
def test_compress():
    import glob

    files = glob.glob(osp.join(here, "hdf5_compression_*.h5"))
    for f in files:
        h = kerchunk.hdf.SingleHdf5ToZarr(
            f, error="raise", unsupported_inline_threshold=800
        )
        if "compression_lz4" in f or "compression_bitshuffle" in f:
            with pytest.raises(RuntimeError):
                h.translate()
            continue
        out = h.translate()
        localfs = AsyncFileSystemWrapper(fsspec.filesystem("file"))
        store = refs_as_store(out, fs=localfs)
        g = zarr.open(store, zarr_format=2)
        assert np.mean(g["data"]) == 49.5

    for f in files:
        # Small chunk of unsupported can be inlined now
        h = kerchunk.hdf.SingleHdf5ToZarr(
            f, error="raise", unsupported_inline_threshold=801
        )
        out = h.translate()
        localfs = AsyncFileSystemWrapper(fsspec.filesystem("file"))
        store = refs_as_store(out, fs=localfs)
        g = zarr.open(store, zarr_format=2)
        assert np.mean(g["data"]) == 49.5


# def test_embed():
#     fn = osp.join(here, "NEONDSTowerTemperatureData.hdf5")
#     h = kerchunk.hdf.SingleHdf5ToZarr(fn, vlen_encode="embed", error="pdb")
#     out = h.translate()
#
#     store = refs_as_store(out)
#     z = zarr.open(store, zarr_format=2)
#     data = z["Domain_10"]["STER"]["min_1"]["boom_1"]["temperature"][:]
#     assert data[0].tolist() == [
#         "2014-04-01 00:00:00.0",
#         "60",
#         "6.72064364129017",
#         "6.667845743708792",
#         "6.774491093631761",
#         "0.0012746926446369846",
#         "0.004609216572327277",
#         "0.01298182345556785",
#     ]
#


def test_inline_threshold():
    fn = osp.join(here, "air.nc")
    inline_0 = kerchunk.hdf.SingleHdf5ToZarr(fn, inline_threshold=0).translate()
    inline_1_million = kerchunk.hdf.SingleHdf5ToZarr(
        fn, inline_threshold=1e9
    ).translate()
    assert inline_0 != inline_1_million


@pytest.mark.skipif(
    not has_visititems_links(),
    reason="'h5py.Group.visititems_links' requires h5py 3.11.0 or later",
)
def test_translate_links():
    fn = osp.join(here, "air_linked.nc")
    # choose a threshold that will give both inline and non-inline
    # datasets for maximum test coverage
    out = kerchunk.hdf.SingleHdf5ToZarr(fn, inline_threshold=50).translate(
        preserve_linked_dsets=True
    )
    localfs = AsyncFileSystemWrapper(fsspec.filesystem("file"))
    store = refs_as_store(out, fs=localfs)
    z = zarr.open(store, zarr_format=2)

    # 1. Test the hard linked datasets were translated correctly
    # 2. Test the soft linked datasets were translated correctly
    for link in ("hard", "soft"):
        for dset in ("lat", "time"):
            np.testing.assert_allclose(z[dset], z[f"{dset}_{link}"])
            for key in z[f"{dset}_{link}"].attrs.keys():
                if key not in kerchunk.hdf._HIDDEN_ATTRS and key != "_ARRAY_DIMENSIONS":
                    assert z[f"{dset}_{link}"].attrs[key] == z[dset].attrs[key]


def test_small_chunks():
    from fsspec.implementations.reference import ReferenceFileSystem

    suffixes = ["blosc", "shuffle_zstd", "zstd", "lz4", "shuffle_blosc"]
    for suffix in suffixes:
        fname = osp.join(here, f"hdf5_mini_{suffix}.h5")
        with open(fname, "rb") as f:
            ref = kerchunk.hdf.SingleHdf5ToZarr(f, None).translate()
        store = ReferenceFileSystem(ref, target=fname, asynchronous=True).get_mapper()
        data = zarr.group(store)["data"][:]
        assert (data == np.arange(6, dtype=np.int32).reshape((3, 2))).all()

        if suffix == "lz4":
            with pytest.raises(RuntimeError):
                with open(fname, "rb") as f:
                    ref = kerchunk.hdf.SingleHdf5ToZarr(
                        f, None, unsupported_inline_threshold=0, error="raise"
                    ).translate()
            continue
        with open(fname, "rb") as f:
            ref = kerchunk.hdf.SingleHdf5ToZarr(
                f, None, unsupported_inline_threshold=0
            ).translate()
        store2 = ReferenceFileSystem(ref, target=fname).get_mapper()
        data2 = zarr.group(store2)["data"][:]
        assert (data2 == np.arange(6, dtype=np.int32).reshape((3, 2))).all()


def test_malicious_chunks():
    from fsspec.implementations.reference import ReferenceFileSystem

    suffixes = ["", "2"]
    for suffix in suffixes:
        fname = osp.join(here, f"hdf5_mali_chunk{suffix}.h5")
        with open(fname, "rb") as f:
            ref = kerchunk.hdf.SingleHdf5ToZarr(f, None).translate()
        store = ReferenceFileSystem(ref, target=fname, asynchronous=True).get_mapper()
        data = zarr.group(store)["data"][:]
        assert (data == np.arange(8, dtype=np.int32)).all()



import os.path

import zarr

import kerchunk.hdf4
from kerchunk.utils import refs_as_store


def test1():
    here = os.path.dirname(__file__)
    fn = os.path.join(here, "MOD14.hdf4")

    out = kerchunk.hdf4.HDF4ToZarr(fn).translate()
    store = refs_as_store(out)
    g = zarr.open(store, zarr_format=2)
    assert g["fire mask"][:].max() == 5



import os.path
import fsspec
import pytest

from kerchunk.utils import refs_as_store


fits = pytest.importorskip("astropy.io.fits")
import kerchunk.fits
import zarr

testdir = os.path.dirname(fits.util.get_testdata_filepath("btable.fits"))
btable = os.path.join(testdir, "btable.fits")
range_im = os.path.join(testdir, "arange.fits")
var = os.path.join(testdir, "variable_length_table.fits")


def test_image():
    # this one directly hits a remote server - should cache?
    url = "https://fits.gsfc.nasa.gov/samples/WFPC2ASSNu5780205bx.fits"
    out = kerchunk.fits.process_file(url)
    g = zarr.open(
        "reference://",
        storage_options=dict(
            fo=out,
            remote_protocol="https",
            asynchronous=True,
            remote_options={"asynchronous": True},
        ),
    )

    # store = refs_as_store(out, remote_protocol="https")
    # g = zarr.open(store)
    arr = g["PRIMARY"][:]
    with fsspec.open(
        "https://fits.gsfc.nasa.gov/samples/WFPC2ASSNu5780205bx.fits"
    ) as f:
        hdu = fits.getdata(f)
        assert (hdu == arr).all()


def test_ascii_table():
    # this one directly hits a remote server - should cache?
    url = "https://fits.gsfc.nasa.gov/samples/WFPC2u5780205r_c0fx.fits"
    out = kerchunk.fits.process_file(url, extension=1)
    store = refs_as_store(out, remote_protocol="https")
    g = zarr.open(store, zarr_format=2)
    arr = g["u5780205r_cvt.c0h.tab"][:]
    with fsspec.open(
        "https://fits.gsfc.nasa.gov/samples/WFPC2u5780205r_c0fx.fits"
    ) as f:
        hdul = fits.open(f)
        hdu = hdul[1]
        assert list(hdu.data.astype(arr.dtype) == arr) == [True, True, True, True]


def test_binary_table():
    out = kerchunk.fits.process_file(btable, extension=1)
    store = refs_as_store(out)
    z = zarr.open(store, zarr_format=2)
    arr = z["1"]
    with open(btable, "rb") as f:
        hdul = fits.open(f)
        attr2 = dict(arr.attrs)
        assert attr2.pop("_ARRAY_DIMENSIONS") == ["x"]
        assert attr2 == dict(hdul[1].header)
        assert (arr[:]["order"] == hdul[1].data["order"]).all()
        assert (arr[:]["mag"] == hdul[1].data["mag"]).all()
        assert (
            arr[:]["name"].astype("U") == hdul[1].data["name"]
        ).all()  # string come out as bytes


def test_cube():
    out = kerchunk.fits.process_file(range_im)
    store = refs_as_store(out)
    z = zarr.open(store, zarr_format=2)
    arr = z["PRIMARY"]
    with open(range_im, "rb") as f:
        hdul = fits.open(f)
        expected = hdul[0].data
    assert (arr[:] == expected).all()


def test_with_class():
    ftz = kerchunk.fits.FitsToZarr(range_im)
    out = ftz.translate()
    assert "fits" in repr(ftz)
    store = refs_as_store(out)
    z = zarr.open(store, zarr_format=2)
    arr = z["PRIMARY"]
    with open(range_im, "rb") as f:
        hdul = fits.open(f)
        expected = hdul[0].data
    assert (arr[:] == expected).all()


def test_var():
    data = fits.open(var)[1].data
    expected = [_.tolist() for _ in data["var"]]

    ftz = kerchunk.fits.FitsToZarr(var)
    out = ftz.translate()
    store = refs_as_store(out)
    z = zarr.open(store, zarr_format=2)
    arr = z["1"]
    vars = [_.tolist() for _ in arr[:]["var"]]

    assert vars == expected
    assert (z["1"][:]["xyz"] == data["xyz"]).all()



from functools import reduce
from operator import mul

import numpy as np
from fsspec.implementations.reference import LazyReferenceMapper
import fsspec

import kerchunk.utils
from kerchunk.utils import (
    _encode_for_JSON,
    dict_to_store,
    inline_array,
    translate_refs_serializable,
)

try:
    from scipy.io._netcdf import ZERO, NC_VARIABLE, netcdf_file, netcdf_variable
except ModuleNotFoundError:  # pragma: no cover
    raise ImportError(
        "Scipy is required for kerchunking NetCDF3 files. Please install with "
        "`pip/conda install scipy`. See https://scipy.org/install/ for more details."
    )


class NetCDF3ToZarr(netcdf_file):
    """Generate references for a netCDF3 file

    Uses scipy's netCDF3 reader, but only reads the metadata. Note that instances
    do behave like actual scipy netcdf files, but contain no valid data.
    Also appears to work for netCDF2, although this is not currently tested.
    """

    def __init__(
        self,
        filename,
        storage_options=None,
        inline_threshold=100,
        max_chunk_size=0,
        out=None,
        **kwargs,
    ):
        """
        Parameters
        ----------
        filename: str
            location of the input
        storage_options: dict
            passed to fsspec when opening filename
        inline_threshold: int
            Byte size below which an array will be embedded in the output. Use 0
            to disable inlining.
        max_chunk_size: int
            How big a chunk can be before triggering subchunking. If 0, there is no
            subchunking, and there is never subchunking for coordinate/dimension arrays.
            E.g., if an array contains 10,000bytes, and this value is 6000, there will
            be two output chunks, split on the biggest available dimension. [TBC]
        out: dict-like or None
            This allows you to supply an fsspec.implementations.reference.LazyReferenceMapper
            to write out parquet as the references get filled, or some other dictionary-like class
            to customise how references get stored
        args, kwargs: passed to scipy superclass ``scipy.io.netcdf.netcdf_file``
        """
        assert kwargs.pop("mmap", False) is False
        assert kwargs.pop("mode", "r") == "r"
        assert kwargs.pop("maskandscale", False) is False

        # attributes set before super().__init__ don't accidentally turn into
        # dataset attributes
        self.chunks = {}
        self.threshold = inline_threshold
        self.max_chunk_size = max_chunk_size
        self.out = out or {}
        self.storage_options = storage_options
        self.fp = fsspec.open(filename, **(storage_options or {})).open()
        magic = self.fp.read(4)
        assert magic[:3] == b"CDF"
        version = kwargs.pop("version", None) or magic[3]
        self.fp.seek(0)
        super().__init__(
            self.fp,
            mmap=False,
            mode="r",
            maskandscale=False,
            version=version,
        )
        self.filename = filename  # this becomes an attribute, so must ignore on write

    def _read_var_array(self):
        header = self.fp.read(4)
        if header not in [ZERO, NC_VARIABLE]:
            raise ValueError("Unexpected header.")

        begin = 0
        dtypes = {"names": [], "formats": []}
        rec_vars = []
        count = self._unpack_int()
        for var in range(count):
            (
                name,
                dimensions,
                shape,
                attributes,
                typecode,
                size,
                dtype_,
                begin_,
                vsize,
            ) = self._read_var()
            if shape and shape[0] is None:  # record variable
                rec_vars.append(name)
                # The netCDF "record size" is calculated as the sum of
                # the vsize's of all the record variables.
                self.__dict__["_recsize"] += vsize
                if begin == 0:
                    begin = begin_
                dtypes["names"].append(name)
                dtypes["formats"].append(str(shape[1:]) + dtype_)

                # Handle padding with a virtual variable.
                if typecode in "bch":
                    actual_size = reduce(mul, (1,) + shape[1:]) * size
                    padding = -actual_size % 4
                    if padding:
                        dtypes["names"].append("_padding_%d" % var)
                        dtypes["formats"].append("(%d,)>b" % padding)

                # Data will be set later.
                data = None
            else:  # not a record variable
                # Calculate size to avoid problems with vsize (above)
                a_size = reduce(mul, shape, 1) * size
                self.chunks[name] = [begin_, a_size, dtype_, shape]
                if name in ["latitude", "longitude", "time"]:
                    pos = self.fp.tell()
                    self.fp.seek(begin_)
                    data = np.frombuffer(self.fp.read(a_size), dtype=dtype_).copy()
                    # data.shape = shape
                    self.fp.seek(pos)
                else:
                    data = np.empty(1, dtype=dtype_)

            # Add variable.
            self.variables[name] = netcdf_variable(
                data,
                typecode,
                size,
                shape,
                dimensions,
                attributes,
                maskandscale=self.maskandscale,
            )

        if rec_vars:
            # Remove padding when only one record variable.
            if len(rec_vars) == 1:
                dtypes["names"] = dtypes["names"][:1]
                dtypes["formats"] = dtypes["formats"][:1]

            pos = self.fp.tell()
            self.fp.seek(begin)
            self.chunks.setdefault("record_array", []).append(
                [begin, self._recs * self._recsize, dtypes]
            )
            self.fp.seek(pos)

    def translate(self):
        """
        Produce references dictionary

        Parameters
        ----------
        """
        import zarr

        out = self.out
        store = dict_to_store(out)
        z = zarr.open_group(store, mode="w", zarr_format=2)

        for dim, var in self.variables.items():
            if dim in self.chunks:
                shape = self.chunks[dim][-1]
            elif dim in self.dimensions:
                shape = self.dimensions[dim]
            else:
                # defer record array
                continue
            if isinstance(shape, int):
                shape = (shape,)
            if shape is None or (len(shape) and shape[0] is None):
                # defer record array
                continue
            else:
                # simple array block
                # TODO: chance to sub-chunk
                fill = var._attributes.get("missing_value", None)
                if fill is None:
                    fill = var._attributes.get("_FillValue", None)
                if fill is not None and var.data.dtype.kind == "f":
                    fill = float(fill)
                if fill is not None and var.data.dtype.kind == "i":
                    fill = int(fill)
                arr = z.create_array(
                    name=dim,
                    shape=shape,
                    dtype=var.data.dtype,
                    fill_value=fill,
                    chunks=shape,
                    compressor=None,
                )
                part = ".".join(["0"] * len(shape)) or "0"
                k = f"{dim}/{part}"
                out[k] = [
                    self.filename,
                    int(self.chunks[dim][0]),
                    int(self.chunks[dim][1]),
                ]
            arr.attrs.update(
                {
                    k: v.decode(errors="ignore") if isinstance(v, bytes) else str(v)
                    for k, v in var._attributes.items()
                    if k not in ["_FillValue", "missing_value"]
                }
            )
            for k in ["add_offset", "scale_factor"]:
                if k in arr.attrs:
                    arr.attrs[k] = float(arr.attrs[k])
            arr.attrs["_ARRAY_DIMENSIONS"] = list(var.dimensions)
        if "record_array" in self.chunks:
            # native chunks version (no codec, no options)
            start, size, dt = self.chunks["record_array"][0]
            dt = np.dtype(dt)
            itemsize = sum(dt[_].itemsize for _ in dt.names)
            outer_shape = size // itemsize
            offset = start
            for name in dt.names:
                dtype = dt[name]

                # Skip padding, but increment offset.
                if name.startswith("_padding_"):
                    offset += dtype.itemsize
                    continue

                # the order of the names if fixed and important!
                var = self.variables[name]
                base = dtype.base  # actual dtype
                shape = (outer_shape,) + dtype.shape

                # TODO: avoid this code repeat
                fill = var._attributes.get("missing_value", None)
                if fill is None:
                    fill = var._attributes.get("_FillValue", None)
                if fill is not None and base.kind == "f":
                    fill = float(fill)
                if fill is not None and base.kind == "i":
                    fill = int(fill)
                arr = z.create_array(
                    name=name,
                    shape=shape,
                    dtype=base,
                    fill_value=fill,
                    chunks=(1,) + dtype.shape,
                    compressor=None,
                    overwrite=True,
                )
                arr.attrs.update(
                    {
                        k: v.decode() if isinstance(v, bytes) else str(v)
                        for k, v in var._attributes.items()
                        if k not in ["_FillValue", "missing_value"]
                    }
                )
                for k in ["add_offset", "scale_factor"]:
                    if k in arr.attrs:
                        arr.attrs[k] = float(arr.attrs[k])

                arr.attrs["_ARRAY_DIMENSIONS"] = list(var.dimensions)

                suffix = (
                    ("." + ".".join("0" for _ in dtype.shape)) if dtype.shape else ""
                )
                for i in range(outer_shape):
                    out[f"{name}/{i}{suffix}"] = [
                        self.filename,
                        int(offset + i * dt.itemsize),
                        int(dtype.itemsize),
                    ]

                offset += dtype.itemsize
        z.attrs.update(
            {
                k: v.decode() if isinstance(v, bytes) else str(v)
                for k, v in self._attributes.items()
                if k != "filename"  # special "attribute"
            }
        )
        out = kerchunk.utils.translate_refs_serializable(out)
        if self.threshold:
            out = inline_array(
                out,
                self.threshold,
                remote_options=self.storage_options,
            )
        if isinstance(out, LazyReferenceMapper):
            out.flush()
            return out
        else:
            translate_refs_serializable(out)
            out = _encode_for_JSON(out)
            return {"version": 1, "refs": out}


netcdf_recording_file = NetCDF3ToZarr  # old name



import os

import fsspec
import numpy as np
from packaging.version import Version
import pytest
from kerchunk import netCDF3

from fsspec.implementations.asyn_wrapper import AsyncFileSystemWrapper
from kerchunk.utils import refs_as_store

xr = pytest.importorskip("xarray")


has_xarray_2023_8_0 = Version(xr.__version__) >= Version("2023.8.0")


arr = np.random.rand(1, 10, 10)
data = xr.DataArray(
    data=arr.squeeze(),
    dims=["x", "y"],
    name="data",
)
bdata = xr.Dataset({"data": data}, attrs={"attr0": 3}).to_netcdf(
    format="NETCDF3_CLASSIC"
)


def test_one(m):
    m.pipe("data.nc3", bdata)
    h = netCDF3.netcdf_recording_file("memory://data.nc3")
    out = h.translate()

    print(out)

    store = refs_as_store(out)

    ds = xr.open_dataset(
        store,
        engine="zarr",
        backend_kwargs={
            "consolidated": False,
            "zarr_format": 2,
        },
    )
    assert (ds.data == data).all()


@pytest.fixture()
def unlimited_dataset(tmpdir):
    # https://unidata.github.io/netcdf4-python/#creatingopeningclosing-a-netcdf-file
    from netCDF4 import Dataset

    fn = os.path.join(tmpdir, "test.nc")
    rootgrp = Dataset(fn, "w", format="NETCDF3_CLASSIC")
    rootgrp.createDimension("time", None)
    rootgrp.createDimension("lat", 10)
    rootgrp.createDimension("lon", 5)
    rootgrp.createVariable("time", "f8", ("time",))
    # reference time is an unbounded dimension that is a half byte long, so it
    # has padding to line up to take up exactly one byte. It is here to test that
    # kerchunk can handle the padding correctly and read following variables
    # correctly.
    rootgrp.createVariable("reference_time", "h", ("time",))
    rootgrp.title = "testing"
    latitudes = rootgrp.createVariable("lat", "f4", ("lat",))
    longitudes = rootgrp.createVariable("lon", "f4", ("lon",))
    temp = rootgrp.createVariable(
        "temp",
        "f4",
        (
            "time",
            "lat",
            "lon",
        ),
    )
    temp.units = "K"
    latitudes[:] = np.arange(-0.5, 0.5, 0.1)
    longitudes[:] = np.arange(0, 0.5, 0.1)
    for i in range(8):
        temp[i] = np.random.uniform(size=(1, 10, 5))
    rootgrp.close()
    return fn


def test_unlimited(unlimited_dataset):
    fn = unlimited_dataset
    expected = xr.open_dataset(fn, engine="scipy")
    h = netCDF3.NetCDF3ToZarr(fn)
    out = h.translate()

    fs = AsyncFileSystemWrapper(fsspec.filesystem("file"))
    store = refs_as_store(out, fs)

    ds = xr.open_zarr(
        store,
        zarr_format=2,
        consolidated=False,
    )
    assert ds.attrs["title"] == "testing"
    assert ds.temp.attrs["units"] == "K"
    assert (ds.lat.values == expected.lat.values).all()
    assert (ds.lon.values == expected.lon.values).all()
    assert (ds.temp.values == expected.temp.values).all()


@pytest.fixture()
def matching_coordinate_dimension_dataset(tmpdir):
    """Create a dataset with a coordinate dimension that matches the name of a
    variable dimension."""
    # https://unidata.github.io/netcdf4-python/#creatingopeningclosing-a-netcdf-file
    from netCDF4 import Dataset

    fn = os.path.join(tmpdir, "test.nc")
    rootgrp = Dataset(fn, "w", format="NETCDF3_64BIT")
    rootgrp.createDimension("node", 2)
    rootgrp.createDimension("sigma", 2)

    node = rootgrp.createVariable("node", "i4", ("node",))
    sigma = rootgrp.createVariable("sigma", "f8", ("sigma", "node"))

    node[:] = [0, 1]
    for i in range(2):
        sigma[i] = np.random.uniform(size=(2,))

    rootgrp.close()
    return fn


@pytest.mark.skipif(
    not has_xarray_2023_8_0, reason="XArray 2023.08.0 is required for this behavior."
)
def test_matching_coordinate_dimension(matching_coordinate_dimension_dataset):
    fn = matching_coordinate_dimension_dataset
    expected = xr.open_dataset(fn, engine="scipy")
    h = netCDF3.NetCDF3ToZarr(fn)
    out = h.translate()
    ds = xr.open_dataset(
        "reference://",
        engine="zarr",
        backend_kwargs={
            "consolidated": False,
            "storage_options": {"fo": out},
        },
    )
    assert (ds.node.values == expected.node.values).all()
    assert (ds.sigma.values == expected.sigma.values).all()
